# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import pytest

from ..version import Token, TokenizedString, parse_version


def reverse_fixtures(table):
    """ Utility method to reverse a list of list. """
    return list(map(list, map(reversed, table)))


@pytest.mark.parametrize('value', [
    0,
    123,
    '0',
    '123',
    'abc',
    'ABC',
    '123abc'])
def test_token_allowed_instanciation(value):
    Token(value)


@pytest.mark.parametrize('value', [
    None,
    -1,
    'a-b-c',
    1.0])
def test_token_unauthorized_instanciation(value):
    with pytest.raises(TypeError):
        Token(value)


eq_values = [
    [Token('2345'), Token('2345')],
    [Token('2345'), Token(2345)],
    [Token('02345'), Token(2345)],
    [Token('02345'), '02345'],
    [Token('2345'), 2345],
    [Token('2345'), '2345'],
    [Token('0'), '0'],
    [Token('0'), 0],
    [Token('000'), 0],
    [Token('abc'), 'abc'],
    [Token('abc'), Token('abc')]]


ne_values = [
    [Token('2345'), Token('45')],
    [Token('2345'), Token(45)],
    [Token('2345'), Token(0)],
    [Token('2345'), Token('0')],
    [Token('2'), -2],
    [Token('2'), 0],
    [Token('abc'), 'def'],
    [Token('Z'), 'z'],
    [Token('acb'), 123]]


gt_values = [
    [Token('9999'), Token('12')],
    [Token('9999'), Token('000000099')],
    [Token(9999), Token('12')],
    [Token(9999), Token(12)],
    [Token(9999), '0'],
    [Token(9999), 0],
    [Token(0), -2],
    [Token('0'), -2],
    [Token('abc'), -2],
    [Token('abc'), 'ab'],
    [Token('a'), 'Z'],
    [Token('z'), 'Z'],
    [Token('a'), Token('Z')]]


lt_values = [
    [Token('12'), Token('9999')],
    [Token('0000012'), Token('9999')],
    [Token('12'), Token(9999)],
    [Token(12), Token(9999)],
    [Token('12'), 9999],
    [Token('0'), 9999],
    [Token('ab'), 'abc'],
    [Token('Z'), 'z']]


@pytest.mark.parametrize('token,value', eq_values)
def test_token_eq(token, value):
    assert token == value


@pytest.mark.parametrize('token,value', ne_values)
def test_token_ne(token, value):
    assert token != value


@pytest.mark.parametrize('token,value', gt_values)
def test_token_gt(token, value):
    assert token > value


@pytest.mark.parametrize('token,value', lt_values)
def test_token_lt(token, value):
    assert token < value


@pytest.mark.parametrize('token,value', gt_values + eq_values)
def test_token_ge(token, value):
    assert token >= value


@pytest.mark.parametrize('token,value', lt_values + eq_values)
def test_token_le(token, value):
    assert token <= value


def test_token_hash():
    assert hash(Token('9999')) == hash(Token('9999'))
    assert hash(Token('9999')) == hash(Token(9999))
    assert hash(Token(9999)) == hash(Token(9999))
    assert hash(Token('09999')) != hash(Token('9999'))
    assert hash(Token('09999')) != hash(Token(9999))


@pytest.mark.parametrize('value', [
    0,
    123,
    -1,
    '0',
    '1.2.3',
    'abc',
    'A-B-C',
    '123   a bc \n'])
def test_tokenized_string_allowed_instanciation(value):
    TokenizedString(value)


@pytest.mark.parametrize('value', [
    None,
    1.0,
    [1, 2, 3]])
def test_tokenized_string_unauthorized_instanciation(value):
    with pytest.raises(TypeError):
        TokenizedString(value)


def test_tokenized_string_hash():
    assert hash(TokenizedString('1.2.3')) == hash(TokenizedString('1.2.3'))
    assert hash(TokenizedString(9999)) == hash(TokenizedString(9999))
    assert hash(TokenizedString('9999')) == hash(TokenizedString(9999))
    assert hash(TokenizedString('09999')) != hash(TokenizedString('9999'))
    assert hash(TokenizedString('09999')) != hash(TokenizedString(9999))
    assert hash(TokenizedString('1.2.3', separator='.')) \
        != hash(TokenizedString('1.2.3', separator='_'))


def test_tokenized_string_idempotent_instanciation():
    tok1 = TokenizedString('1.2.3')
    tok2 = TokenizedString(tok1)
    assert tok1 is tok2
    assert hash(tok1) == hash(tok2)
    assert tok1 == tok2


version_list = [
    ['r2917_1', ('r', 2917, 1)],
    [' r     29   \n  17   1 x  ', ('r', 29, 17, 1, 'x')],
    ['2020.03.24', (2020, 3, 24)],
    ['4.2.1-5666.3', (4, 2, 1, 5666, 3)],
    ['4.2.1-5666..3', (4, 2, 1, 5666, 3)],
    ['4.2.1-5666.3.', (4, 2, 1, 5666, 3)],
    ['.4.2.1-5666.3', (4, 2, 1, 5666, 3)],
    ['4.2.1--5666.3', (4, 2, 1, 5666, 3)],
    ['4.2.1-#-5666.3', (4, 2, 1, 5666, 3)],
    ['4.2.1-ABC-5666.3', (4, 2, 1, 'abc', 5666, 3)],
    ['4.2.1-éèàçÇÉÈ²³¼ÀÁÂÃÄ—ÅËÍÑÒÖÜÝåïš™-5666.3',
     (4, 2, 1, 'eeaccee', 231, 4, 'aaaaae', 'ainooeueyastm', 5666, 3)],
    ['1.3_1', (1, 3, 1)],
    ['2.40.20161221.0239', (2, 40, 20161221, 239)],
    ['latest', ('latest',)],
    ['1.2beta5', (1, 2, 'beta', 5)],
    ['2.8.18-x86_64', (2, 8, 18, 'x', 86, 64)],
    ['1.8.0_112-b16', (1, 8, 0, 112, 'b', 16)],
    ['3.6.9_build_4685', (3, 6, 9, 'build', 4685)],
    ['1.8.6-124-g6cd4c31', (1, 8, 6, 124, 'g', 6, 'cd', 4, 'c', 31)],
    ['0.9999999', (0, 9999999)],
    ['0.0.0', (0, 0, 0)],
    ['1.0+git71+c79e264-r0', (1, 0, 'git', 71, 'c', 79, 'e', 264, 'r', 0)],
    ['20190718-r0', (20190718, 'r', 0)],
    ['0.0.0-development', (0, 0, 0, 'development')],
    ['v0.9.4.3', ('v', 0, 9, 4, 3)],
    ['6.0.1-0ubuntu1', (6, 0, 1, 0, 'ubuntu', 1)],
    ['4.6.0+git+20160126-2', (4, 6, 0, 'git', 20160126, 2)],
    ['1:5.25-2ubuntu1', (1, 5, 25, 2, 'ubuntu', 1)],
    ['1.18.4ubuntu1.1', (1, 18, 4, 'ubuntu', 1, 1)],
    ['20160104ubuntu1', (20160104, 'ubuntu', 1)],
    ['3.113+nmu3ubuntu4', (3, 113, 'nmu', 3, 'ubuntu', 4)],
    ['13.0.2,8:d4173c853231432f001e99d882',
     (13, 0, 2, 8, 'd', 4173, 'c', 853231432, 'f', 1, 'e', 99, 'd', 882)],
    ['1.01+20150706hgc3698e0+dfsg-2',
     (1, 1, 20150706, 'hgc', 3698, 'e', 0, 'dfsg', 2)]]


@pytest.mark.parametrize('v_string,v_tuple', version_list)
def test_version_tokenizer(v_string, v_tuple):
    assert TokenizedString(v_string) == v_tuple


compared_gt = [
    ['2.0', '1.0'],
    ['0.1', '0'],
    ['0.1', '0.0'],
    ['0.1', '0.0.0.0.0'],
    ['2.0', '1.0'],
    ['2.0', '1'],
    ['3.1.10', '3.1.9'],
    ['8.2p1_1', '8.2p1'],
    [20200313, 20190801],
    ['2020.03.24', '2019.11.28']]


@pytest.mark.parametrize('ver1,ver2', compared_gt)
def test_version_comparison_gt(ver1, ver2):
    assert TokenizedString(ver1) > TokenizedString(ver2)
    assert parse_version(ver1) > parse_version(ver2)


@pytest.mark.parametrize('ver1,ver2', reverse_fixtures(compared_gt))
def test_version_comparison_lt(ver1, ver2):
    assert TokenizedString(ver1) < TokenizedString(ver2)
    assert parse_version(ver1) < parse_version(ver2)



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Utilities to manage GitHub labels to use in issues and PR management.
"""

from boltons.iterutils import flatten

from .managers import pool
from .platform import ALL_OS_LABELS

# Format: label name, color, optional description.
LABELS = [
    ("BitBar plugin", '#fef2c0',
     "Plugin code, documentation or features related to BitBar"),
    ("bug", '#d73a4a',
     "Something isn't working, or a fix is proposed"),
    ("CI/CD", '#dbca13',
     "Tests, automation and management of the project"),
    ("documentation", '#006b75',
     "Update to non-code (manual, readme, tutorial, docstrings, ...) and "
     "its generation"),
    ("duplicate", '#cfd3d7',
     "This issue or pull request already exists"),
    ("enhancement", '#84b6eb',
     "Improvement or change of an existing feature"),
    ("feature request", '#fbca04',
     "Something not existing yet that need to be implemented"),
    ("good first issue", '#7057ff',
     "A place for newcomers to start contributing"),
    ("help wanted", '#008672',
     "Extra attention is needed"),
    ("can't reproduce", '#fec1c1',
     "Root cause unlikely to come from the project"),
    ("question", '#d876e3',
     "Further information is requested"),
    ("wont do/fix", '#eeeeee',
     "This will not be worked on"),
]


# Maps platform and manager ID to their labels.
PLATFORM_LABELS = {}
MANAGER_LABELS = {}


# Some managers sharing some roots will be grouped together.
MANAGER_GROUPS = {
    'dpkg-like': {'dpkg', 'apt', 'opkg'},
    'npm-like': {'npm', 'yarn'},
}

# Define some colors.
PLATFORM_COLOR = '#bfd4f2'
MANAGER_COLOR = '#bfdadc'


# Create one label per platform.
for platform_id in ALL_OS_LABELS:
    label_id = 'platform: {}'.format(platform_id)
    LABELS.append((label_id, PLATFORM_COLOR, platform_id))
    PLATFORM_LABELS[platform_id] = label_id


# Create one label per manager. Add mpm as its own manager.
non_grouped_managers = set(
    pool()) - set(flatten(MANAGER_GROUPS.values())) | {'mpm'}
for manager_id in non_grouped_managers:
    label_id = 'manager: {}'.format(manager_id)
    LABELS.append((label_id, MANAGER_COLOR, manager_id))
    if manager_id != 'mpm':
        MANAGER_LABELS[manager_id] = label_id


# Add labels for grouped managers.
for group_label, manager_ids in MANAGER_GROUPS.items():
    label_id = 'manager: {}'.format(group_label)
    LABELS.append((label_id, MANAGER_COLOR, ', '.join(sorted(manager_ids))))
    for manager_id in manager_ids:
        MANAGER_LABELS[manager_id] = label_id



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import os

import pytest

from .conftest import destructive, run_cmd, unless_macos

# Default location of Homebrew Cask formulas on macOS. This is supposed to be a
# shallow copy of the following Git repository:
# https://github.com/Homebrew/homebrew-cask
CASK_PATH = '/usr/local/Homebrew/Library/Taps/homebrew/homebrew-cask/Casks/'


@pytest.fixture
def install_cask():

    packages = set()

    def git_checkout(package_id, commit):
        code, _, _ = run_cmd(
            'git', '-C', CASK_PATH, 'checkout', commit,
            "{}.rb".format(package_id))
        assert code == 0

    def _install_cask(package_id, commit):
        packages.add(package_id)
        # Deepen homebrew repository copy so we can dig into the past.
        # Arbitrary set oldest reference to 2018-01-01, which gives us enough
        # to dig into the past.
        code, _, _ = run_cmd(
            'git', '-C', CASK_PATH, 'fetch', '--shallow-since=2018-01-01')
        assert code == 0
        # Fetch locally the old version of the Cask's formula.
        git_checkout(package_id, commit)
        # Install the cask but bypass its local cache auto-update: we want to
        # force brew to rely on our formula from the past.
        os.environ['HOMEBREW_NO_AUTO_UPDATE'] = '1'
        code, output, error = run_cmd('brew', 'cask', 'reinstall', package_id)
        # Reset our temporary environment variable.
        del os.environ['HOMEBREW_NO_AUTO_UPDATE']
        # Restore old formula to its most recent version.
        git_checkout(package_id, 'HEAD')
        # Check the cask has been properly installed.
        assert code == 0
        if error:
            assert "is already installed" not in error
        assert "{} was successfully installed!".format(package_id) in output
        return output

    yield _install_cask

    # Remove all installed packages.
    for package_id in packages:
        run_cmd('brew', 'cask', 'uninstall', '--force', package_id)


@destructive
@unless_macos
class TestCask:

    def test_autoupdate_unicode_name(self, invoke, install_cask):
        """ See #16. """
        # Install an old version of a package with a unicode name.
        # Old Cask formula for ubersicht 1.4.60.
        output = install_cask(
            'ubersicht', 'bb72da6c085c017f6bccebbfee5e3bc4837f3165')
        assert 'Uebersicht-1.4.60.app.zip' in output
        assert 'Übersicht.app' in output
        assert 'Übersicht.app' not in output

        # Ubersicht is not reported as outdated because is tagged as
        # auto-update.
        result = invoke('--manager', 'cask', 'outdated')
        assert result.exit_code == 0
        assert "ubersicht" not in result.stdout
        assert "Übersicht" not in result.stdout

        # Try with explicit option.
        result = invoke(
            '--ignore-auto-updates', '--manager', 'cask', 'outdated')
        assert result.exit_code == 0
        assert "ubersicht" not in result.stdout
        assert "Übersicht" not in result.stdout

        # Look for reported available upgrade.
        result = invoke(
            '--include-auto-updates', '--manager', 'cask', 'outdated')
        assert result.exit_code == 0
        assert "ubersicht" in result.stdout
        # Outdated subcommand does not fetch the unicode name by default.
        assert "Übersicht" not in result.stdout

    def test_multiple_names(self, invoke, install_cask):
        """ See #26. """
        # Install an old version of a package with multiple names.
        # Old Cask formula for xld 2018.10.19.
        output = install_cask(
            'xld', '89536da7075aa3ac9683a67189fddbed4a7d818c')
        assert 'xld-20181019.dmg' in output
        assert 'XLD.app' in output

        # Look for reported available upgrade.
        result = invoke(
            '--include-auto-updates', '--manager', 'cask', 'outdated')
        assert result.exit_code == 0
        assert "xld" in result.stdout
        # Outdated subcommand does not fetch the unicode name by default.
        assert "X Lossless Decoder" not in result.stdout



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import textwrap

import pytest

from .conftest import MANAGER_IDS
from .test_cli import CLISubCommandTests


def create_toml(filename, content):
    """ Utility to produce TOML files. """
    with open(filename, 'w') as doc:
        doc.write(textwrap.dedent(content.strip()))


@pytest.fixture
def subcmd():
    """ Seed common subcommand tests with a dummy file and content to allow the
    CLI to not fail on required file input. """
    create_toml('dummy.toml', """
        [dummy_manager]
        fancy_package = "0.0.1"
        """)
    return 'restore', 'dummy.toml'


class TestRestore(CLISubCommandTests):

    def test_default_all_managers(self, invoke):
        create_toml('all-managers.toml', ''.join(["""
            [{}]
            blah = 123
            """.format(m) for m in MANAGER_IDS]))

        result = invoke('restore', 'all-managers.toml')
        assert result.exit_code == 0
        assert 'all-managers.toml' in result.stderr
        self.check_manager_selection(result)

    @pytest.mark.parametrize('mid', MANAGER_IDS)
    def test_single_manager(self, invoke, mid):
        create_toml('all-managers.toml', ''.join(["""
            [{}]
            blah = 123
            """.format(m) for m in MANAGER_IDS]))

        result = invoke('--manager', mid, 'restore', 'all-managers.toml')
        assert result.exit_code == 0
        self.check_manager_selection(result, {mid})

    def test_ignore_unrecognized_manager(self, invoke):
        create_toml('unrecognized.toml', """
            [random_section]
            blah = 123
            """)

        result = invoke('restore', 'unrecognized.toml')
        assert result.exit_code == 0
        assert 'unrecognized.toml' in result.stderr
        assert 'warning: Ignore [random_section] section.' in result.stderr

    def test_restore_single_manager(self, invoke):
        create_toml('pip-npm-dummy.toml', """
            [pip]
            fancy_package = "0.0.1"

            [npm]
            dummy_package = "2.2.2"
            """)

        result = invoke('--manager', 'npm', 'restore', 'pip-npm-dummy.toml')
        assert result.exit_code == 0
        assert 'pip-npm-dummy.toml' in result.stderr
        assert 'Restore pip' not in result.stderr
        assert 'Restore npm' in result.stderr

    def test_restore_excluded_manager(self, invoke):
        create_toml('pip-npm-dummy.toml', """
            [pip]
            fancy_package = "0.0.1"

            [npm]
            dummy_package = "2.2.2"
            """)

        result = invoke('--exclude', 'npm', 'restore', 'pip-npm-dummy.toml')
        assert result.exit_code == 0
        assert 'pip-npm-dummy.toml' in result.stderr
        assert 'Restore pip' in result.stderr
        assert 'Restore npm' not in result.stderr

    def test_empty_manager(self, invoke):
        create_toml('pip-empty.toml', """
            [pip]
            """)

        result = invoke('restore', 'pip-empty.toml')
        assert result.exit_code == 0
        assert 'pip-empty.toml' in result.stderr
        assert 'Restore pip' in result.stderr



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import pytest
import simplejson as json

from .conftest import MANAGER_IDS
from .test_cli import CLISubCommandTests, CLITableTests


@pytest.fixture
def subcmd():
    return 'outdated'


BITBAR_KEYWORDS = {
   'bash',
}.union({"param{}".format(i) for i in range(1, 10)})


class TestOutdated(CLISubCommandTests, CLITableTests):

    @pytest.mark.parametrize('mid', MANAGER_IDS)
    def test_single_manager(self, invoke, mid, subcmd):
        result = invoke('--manager', mid, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, {mid})

    def test_json_parsing(self, invoke, subcmd):
        result = invoke('--output-format', 'json', subcmd)
        assert result.exit_code == 0
        data = json.loads(result.stdout)

        assert data
        assert isinstance(data, dict)
        assert set(data).issubset(MANAGER_IDS)

        for manager_id, info in data.items():
            assert isinstance(manager_id, str)
            assert isinstance(info, dict)

            assert isinstance(info['id'], str)
            assert isinstance(info['name'], str)

            keys = {'errors', 'id', 'name', 'packages'}
            if 'upgrade_all_cli' in info:
                assert isinstance(info['upgrade_all_cli'], str)
                keys.add('upgrade_all_cli')
            assert set(info) == keys

            assert isinstance(info['errors'], list)
            if info['errors']:
                assert set(map(type, info['errors'])) == {str}

            assert info['id'] == manager_id

            assert isinstance(info['packages'], list)
            for pkg in info['packages']:
                assert isinstance(pkg, dict)

                assert set(pkg) == {
                    'id', 'installed_version', 'latest_version', 'name',
                    'upgrade_cli'}

                assert isinstance(pkg['id'], str)
                assert isinstance(pkg['installed_version'], str)
                assert isinstance(pkg['latest_version'], str)
                assert isinstance(pkg['name'], str)
                assert isinstance(pkg['upgrade_cli'], str)

    def test_cli_format_plain(self, invoke, subcmd):
        result = invoke(
            '--output-format', 'json', subcmd, '--cli-format', 'plain')
        for outdated in json.loads(result.stdout).values():
            for infos in outdated['packages']:
                assert isinstance(infos['upgrade_cli'], str)

    def test_cli_format_fragments(self, invoke, subcmd):
        result = invoke(
            '--output-format', 'json', subcmd, '--cli-format', 'fragments')
        for outdated in json.loads(result.stdout).values():
            for infos in outdated['packages']:
                assert isinstance(infos['upgrade_cli'], list)
                assert set(map(type, infos['upgrade_cli'])) == {str}

    def test_cli_format_bitbar(self, invoke, subcmd):
        result = invoke(
            '--output-format', 'json', subcmd, '--cli-format', 'bitbar')
        for outdated in json.loads(result.stdout).values():
            for infos in outdated['packages']:
                assert isinstance(infos['upgrade_cli'], str)
                assert 'param1=' in infos['upgrade_cli']
                for param in infos['upgrade_cli'].split(' '):
                    k, v = param.split('=', 1)
                    assert k in BITBAR_KEYWORDS
                    assert set(v.lower()).issubset(
                        '0123456789abcdefghijklmnopqrstuvwxyz./-_+="\\@:')



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re
from collections import Counter
from pathlib import Path

import pytest
import simplejson as json
from boltons.iterutils import flatten
from yaml import Loader, load

from ..labels import MANAGER_LABELS, PLATFORM_LABELS
from ..managers import pool
from ..platform import OS_DEFINITIONS, os_label
from .conftest import MANAGER_IDS

""" Test all non-code artifacts depending on manager definitions.

Covers:
    * Documentation (sphinx, readme, etc.)
    * CI/CD scripts
    * GitHub project config

These tests are mainly there to remind us keep extra stuff in sync on new
platform or manager addition.
"""

PROJECT_ROOT = Path(__file__).parent.parent.parent


def test_changelog():
    with PROJECT_ROOT.joinpath('CHANGES.rst').resolve().open() as doc:
        content = doc.read()

    assert content.startswith("Changelog\n=========\n")

    entry_pattern = re.compile(r"^\* \[(?P<category>[a-z,]+)\] (?P<entry>.+)")
    for line in content.splitlines():
        if line.startswith('*'):
            match = entry_pattern.match(line)
            assert match
            entry = match.groupdict()
            assert entry['category']
            assert set(entry['category'].split(',')).issubset(flatten([
                MANAGER_IDS, OS_DEFINITIONS.keys(), 'mpm', 'bitbar']))


def test_labeller_rules():

    # Extract list of canonically defined labels.
    with PROJECT_ROOT.joinpath('.github/labels.json').resolve().open() as doc:
        content = doc.read()
    defined_labels = [lbl['name'] for lbl in json.loads(content)]

    # Canonical labels are uniques.
    assert len(defined_labels) == len(set(defined_labels))
    canonical_labels = set(defined_labels)

    # Extract and categorize labels.
    canonical_managers = {lbl for lbl in canonical_labels
                          if lbl.startswith('manager: ') and 'mpm' not in lbl}
    canonical_platforms = {lbl for lbl in canonical_labels
                           if lbl.startswith('platform: ')}
    assert canonical_managers
    assert canonical_platforms

    # Extract rules from json blurb serialized into YAML.
    with PROJECT_ROOT.joinpath(
            '.github/workflows/labels_issue.yaml').resolve().open() as doc:
        content = doc.read()
    assert "Naturalclar/issue-action" in content
    json_rules = load(content, Loader=Loader)[
        'jobs']['labeller']['steps'][0]['with']['parameters']
    rules = json.loads(json_rules)
    assert rules

    # Each keyword match one rule only.
    rules_keywords = Counter(flatten([r['keywords'] for r in rules]))
    assert rules_keywords
    assert max(rules_keywords.values()) == 1

    # Extract and categorize labels.
    rules_labels = Counter(flatten([r['labels'] for r in rules]))

    assert rules_labels
    # Check that all labels are canonically defined.
    assert canonical_labels.issuperset(rules_labels)

    rules_managers = Counter({
        label: count for label, count in rules_labels.items()
        if label.startswith('manager: ')})
    rules_platforms = Counter({
        label: count for label, count in rules_labels.items()
        if label.startswith('platform: ')})

    assert rules_managers
    # Each canonical manager labels is defined.
    assert len(canonical_managers.symmetric_difference(rules_managers)) == 0
    # Each manager has a rule and one only.
    assert max(rules_managers.values()) == 1

    assert rules_platforms
    # Each canonical platform labels is defined.
    assert len(canonical_platforms.symmetric_difference(rules_platforms)) == 0
    # Each registered OS has a rule.
    assert len(rules_platforms) == len(OS_DEFINITIONS)
    # Each platforms has at least a rule.
    assert min(rules_platforms.values()) >= 1

    # Check that all labels are canonically defined.
    assert canonical_labels.issuperset(rules_labels)

    # Check each rule definition.
    for rule in rules:

        # No duplicate labels.
        assert len(set(rule['labels'])) == len(rule['labels'])

        # Special checks for rules targetting manager labels.
        manager_label = canonical_managers.intersection(rule['labels'])
        if manager_label:

            # Extract manager label
            assert len(manager_label) == 1
            manager_label = manager_label.pop()

            # Only platforms are expected alongside manager labels.
            platforms = set(rule['labels']) - canonical_managers
            assert platforms.issubset(canonical_platforms)

            # Check managers sharing the same label shares the same platforms.
            supported_platforms = [
                pool()[mid].platforms
                for mid, lbl in MANAGER_LABELS.items()
                # Relying on pool() restrict our checks, as the pool exclude
                # non-locally supported managers.
                if lbl == manager_label and mid in pool()]
            assert len(set(supported_platforms)) == 1

            # Check the right platforms is associated with the manager.
            supported_platform_labels = {
                PLATFORM_LABELS[os_label(p)] for p in supported_platforms[0]}
            assert platforms == supported_platform_labels



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import logging
import re
from pathlib import Path

import click
import pytest
import simplejson as json
from boltons.iterutils import flatten
from boltons.strutils import strip_ansi
from cli_helpers.tabular_output import TabularOutputFormatter

from .. import __version__, logger
from ..cli import RENDERING_MODES, WINDOWS_MODE_BLACKLIST
from .conftest import MANAGER_IDS, unless_windows

""" Common tests for all CLI basic features and templates for subcommands. """


def test_real_fs():
    """ Check a simple test is not caught into the CLI runner fixture which is
    encapsulating all filesystem access into temporary directory structure. """
    assert str(Path(__file__)).startswith(str(Path.cwd()))


def test_temporary_fs(runner):
    """ Check the CLI runner fixture properly encapsulated the filesystem in
    temporary directory. """
    assert not str(Path(__file__)).startswith(str(Path.cwd()))


class TestBaseCLI:

    """ This collection is testing basic CLI behavior shared by all
    subcommands.

    Also regroups tests not involving subcommands.

    Also includes a bunch of tests performed once on an arbitrary sub-command,
    for situation when the tested behavior is shared by all subcommands. In
    that case we choosed `managers` as it is a safe read-only operation
    supposed to work on all platforms whatever the environment.
    """

    def test_bare_call(self, invoke):
        result = invoke()
        assert result.exit_code == 0
        assert "Usage: " in result.stdout
        assert not result.stderr

    def test_main_help(self, invoke):
        result = invoke('--help')
        assert result.exit_code == 0
        assert "Usage: " in result.stdout
        assert not result.stderr

    def test_version(self, invoke):
        result = invoke('--version')
        assert result.exit_code == 0
        assert __version__ in result.stdout
        assert not result.stderr

    def test_unknown_option(self, invoke):
        result = invoke('--blah')
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: no such option: --blah" in result.stderr

    def test_unknown_command(self, invoke):
        result = invoke('blah')
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: No such command 'blah'." in result.stderr

    def test_required_command(self, invoke):
        result = invoke('--verbosity', 'DEBUG')
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: Missing command." in result.stderr

    def test_unrecognized_verbosity(self, invoke):
        result = invoke('--verbosity', 'random', 'managers')
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: Invalid value for '--verbosity' / '-v'" in result.stderr

    @pytest.mark.parametrize(
        'level', ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'])
    def test_verbosity(self, invoke, level):
        result = invoke('--verbosity', level, 'managers')
        assert result.exit_code == 0
        assert logger.level == getattr(logging, level)
        assert "══════" in result.stdout
        if level == 'DEBUG':
            assert "debug: " in result.stderr
        else:
            assert "debug: " not in result.stderr

    @unless_windows
    @pytest.mark.parametrize('mode', WINDOWS_MODE_BLACKLIST)
    def test_check_failing_unicode_rendering(self, mode):
        """ Check internal assumption that some rendering unicode table
        rendering modes fails in Windows console. """
        table_formatter = TabularOutputFormatter(mode)
        with pytest.raises(UnicodeEncodeError):
            click.echo(table_formatter.format_output(
                ((1, 87), (2, 80), (3, 79)), ('day', 'temperature')))


class CLISubCommandTests:

    """ All these tests runs on each subcommand.

    This class doesn't starts with `Test` as it is meant to be used as a
    template, inherited sub-command specific files.
    """

    def test_help(self, invoke, subcmd):
        result = invoke(subcmd, '--help')
        assert result.exit_code == 0
        assert "Usage: " in result.stdout
        assert flatten([subcmd])[0] in result.stdout
        assert not result.stderr

    @pytest.mark.parametrize('opt_stats', ['--stats', '--no-stats', None])
    @pytest.mark.parametrize('opt_timer', ['--time', '--no-time', None])
    def test_options(self, invoke, subcmd, opt_stats, opt_timer):
        """ Test the result on all combinations of optional options. """
        result = invoke(opt_stats, opt_timer, subcmd)
        assert result.exit_code == 0

    @staticmethod
    def check_manager_selection(result, selected=MANAGER_IDS):
        """ Check inclusion and exclusion of a set of managers.

        Check all manager are there by default.

        .. todo:

            Parametrize/fixturize signals to pin point output depending on
            subcommand.
        """
        assert isinstance(selected, (frozenset, set))

        found_managers = set()
        skipped_managers = set()

        for mid in MANAGER_IDS:

            # List of signals indicating a package manager has been retained by
            # the CLI. Roughly sorted from most specific to more forgiving.
            signals = [
                # Common "not found" warning message.
                "warning: Skip unavailable {} manager.".format(
                    mid) in result.stderr,
                # Common "not implemented" optional command warning message.
                bool(re.search(
                    r"warning: (Sync|Cleanup) not implemented for {}.".format(
                        mid), result.stderr)),
                # Stats line at the end of output.
                "{}: ".format(mid) in result.stdout.splitlines(
                    )[-1] if result.stdout else '',
                # Match output of managers command.
                bool(re.search(r"\s+│\s+{}\s+│\s+(✓|✘).+│\s+(✓|✘)".format(mid),
                               strip_ansi(result.stdout))),
                # Sync command.
                "Sync {} package info...".format(mid) in result.stderr,
                # Upgrade command.
                "Updating all outdated packages from {}...".format(
                    mid) in result.stderr,
                # Cleanup command.
                "Cleanup {}...".format(mid) in result.stderr,
                # Log message for backup command.
                "Dumping packages from {}...".format(mid) in result.stderr,
                # Restoring message.
                "Restore {} packages...".format(mid) in result.stderr,
                # Warning message for restore command.
                "warning: No [{}] section found.".format(mid) in result.stderr]

            if True in signals:
                found_managers.add(mid)
            else:
                skipped_managers.add(mid)

        # Compare managers reported by the CLI and those expected.
        assert found_managers == selected
        assert skipped_managers == MANAGER_IDS - selected

    @pytest.mark.parametrize('selector', ['--manager', '--exclude'])
    def test_invalid_manager_selector(self, invoke, subcmd, selector):
        result = invoke(selector, 'unknown', subcmd)
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: Invalid value for " in result.stderr
        assert selector in result.stderr

    def test_default_all_managers(self, invoke, subcmd):
        """ Test all available managers are selected by default. """
        result = invoke(subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result)

    @pytest.mark.parametrize('args,expected', [
        pytest.param(
            ('--manager', 'apm'), {'apm'},
            id="single_selector"),
        pytest.param(
            ('--manager', 'apm') * 2, {'apm'},
            id="duplicate_selectors"),
        pytest.param(
            ('--manager', 'apm', '--manager', 'gem'), {'apm', 'gem'},
            id="multiple_selectors"),
        pytest.param(
            ('--exclude', 'apm'), MANAGER_IDS - {'apm'},
            id="single_exclusion"),
        pytest.param(
            ('--exclude', 'apm') * 2, MANAGER_IDS - {'apm'},
            id="duplicate_exclusions"),
        pytest.param(
            ('--exclude', 'apm', '--exclude', 'gem'),
            MANAGER_IDS - {'apm', 'gem'},
            id="multiple_exclusions"),
        pytest.param(
            ('--manager', 'apm', '--exclude', 'gem'), {'apm'},
            id="priority_ordered"),
        pytest.param(
            ('--exclude', 'gem', '--manager', 'apm'), {'apm'},
            id="priority_reversed"),
        pytest.param(
            ('--manager', 'apm', '--exclude', 'apm'), set(),
            id="exclusion_override_ordered"),
        pytest.param(
            ('--exclude', 'apm', '--manager', 'apm'), set(),
            id="exclusion_override_reversed"),
    ])
    def test_manager_selection(self, invoke, subcmd, args, expected):
        result = invoke(*args, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, expected)


class CLITableTests:

    """ Test subcommands whose output is a configurable table.

    A table output is also allowed to be rendered as JSON.
    """

    # List of all supported rendering modes IDs, and their expected output.
    expected_renderings = {
        'ascii':          ('---+---',               None),
        'csv':            (',',                     None),
        'csv-tab':        ('\t',                    None),
        'double':         ('═══╬═══',               None),
        'fancy_grid':     ('═══╪═══',               None),
        'github':         ('---|---',               None),
        'grid':           ('===+===',               'ascii'),
        'html':           ('<table>',               None),
        'jira':           (' || ',                  None),
        'json':           ('": {',                  None),
        'latex':          ('\\hline',               None),
        'latex_booktabs': ('\\toprule',             None),
        'mediawiki':      ('{| class="wikitable" ', 'jira'),
        'moinmoin':       (" ''' || ''' ",          'jira'),
        'orgtbl':         ('---+---',               None),
        'pipe':           ('---|:---',              None),
        'plain':          ('  ',                    None),
        'psql':           ('---+---',               None),
        'rst':            ('===  ===',              None),
        'simple':         ('---  ---',              None),
        'textile':        (' |_. ',                 None),
        'tsv':            ('\t',                    None),
        'vertical':       ('***[ 1. row ]***',      None)}

    def test_recognized_modes(self):
        """ Check all rendering modes proposed by the table module are
        accounted for. """
        assert RENDERING_MODES == set(self.expected_renderings.keys())

    def test_default_table_rendering(self, invoke, subcmd):
        """ Check default rendering is fancy_grid. """
        result = invoke(subcmd)
        assert result.exit_code == 0

        expected = self.expected_renderings['fancy_grid'][0]

        # If no package found, check that no table gets rendered. Else, check
        # the selected mode is indeed rendered in <stdout>, so the CLI result
        # can be grep-ed.
        if result.stdout.startswith("0 package total ("):
            assert expected not in result.stdout
        else:
            assert expected in result.stdout

        assert expected not in result.stderr

    @pytest.mark.parametrize('mode,expected,conflict', [
        (k, v[0], v[1]) for k, v in expected_renderings.items()])
    def test_all_table_rendering(
            self, invoke, subcmd, mode, expected, conflict):
        """ Check that from all rendering modes, only the selected one appears
        in <stdout> and only there. Any other mode are not expected neither in
        <stdout> or <stderr>.
        """
        result = invoke('--output-format', mode, subcmd)
        assert result.exit_code == 0

        # If no package found, check that no table gets rendered. Else, check
        # the selected mode is indeed rendered in <stdout>, so the CLI result
        # can be grep-ed.
        if result.stdout.startswith("0 package total ("):
            # CSV mode will match on comma.
            if mode != 'csv':
                assert expected not in result.stdout
        else:
            assert expected in result.stdout

        # Collect all possible unique traces from all possible rendering modes.
        blacklist = {
            v[0] for v in self.expected_renderings.values()
            # Exclude obvious character sequences shorter than 3 characters to
            # eliminate false negative.
            if len(v[0]) > 2}
        # Remove overlapping edge-cases.
        if conflict:
            blacklist.remove(self.expected_renderings[conflict][0])

        for unexpected in blacklist:
            if unexpected != expected:
                # The unexpected trace is not the selected one, it should not
                # appears at all in stdout.
                assert unexpected not in result.stdout
            # Any expected output from all rendering modes must not appears in
            # <stderr>, including the selected one.
            assert unexpected not in result.stderr

    def test_json_debug_output(self, invoke, subcmd):
        """ Output is expected to be parseable if read from <stdout> even in
        debug level as these messages are redirected to <stderr>.

        Also checks that JSON output format is not supported by all commands.
        """
        result = invoke(
            '--output-format', 'json', '--verbosity', 'DEBUG', subcmd)
        assert result.exit_code == 0
        assert "debug:" in result.stderr
        json.loads(result.stdout)
        json.loads(result.output)
        with pytest.raises(json.decoder.JSONDecodeError):
            json.loads(result.stderr)



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import pytest
import simplejson as json

from .conftest import MANAGER_IDS
from .test_cli import CLISubCommandTests, CLITableTests


@pytest.fixture
def subcmd():
    return 'managers'


class TestManagers(CLISubCommandTests, CLITableTests):

    @pytest.mark.parametrize('mid', MANAGER_IDS)
    def test_single_manager(self, invoke, subcmd, mid):
        result = invoke('--manager', mid, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, {mid})

    def test_json_parsing(self, invoke, subcmd):
        result = invoke('--output-format', 'json', subcmd)
        assert result.exit_code == 0
        data = json.loads(result.stdout)

        assert data
        assert isinstance(data, dict)
        assert set(data) == MANAGER_IDS

        for manager_id, info in data.items():
            assert isinstance(manager_id, str)
            assert isinstance(info, dict)

            assert set(info) == {
                'available', 'cli_path', 'errors', 'executable', 'fresh', 'id',
                'name', 'supported', 'version'}

            assert isinstance(info['available'], bool)
            if info['cli_path'] is not None:
                assert isinstance(info['cli_path'], str)

            assert isinstance(info['errors'], list)
            if info['errors']:
                assert set(map(type, info['errors'])) == {str}

            assert isinstance(info['executable'], bool)
            assert isinstance(info['fresh'], bool)
            assert isinstance(info['id'], str)
            assert isinstance(info['name'], str)
            assert isinstance(info['supported'], bool)

            if info['version'] is not None:
                assert isinstance(info['version'], str)

            assert info['id'] == manager_id
